/* $Id: e2_file_info_dialog.c 3 2006-11-01 04:36:04Z tpgww $
 *
 * Copyright (C) 2003-2006 tooar <tooar@gmx.net>
 * Portions copyright (C) 1999 Michael Clark.
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "emelfm2.h"
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <pwd.h>
#include <grp.h>
#include "e2_dialog.h"

//max length of the file type string
#define FILE_TYPE_LENGTH 128

//filename is localised
static gboolean _e2_file_info_dialog_get_file_type (gchar *filepath,
	gchar **readable_type, gchar **mime_type)
{
	size_t bsize = strlen (filepath) + 512;
	gchar *buf = (gchar *) g_try_malloc ((gulong) bsize);
	if (buf == NULL)
	{
		*readable_type = NULL;
		*mime_type = NULL;
		return FALSE;
	}
	//CHECKME -b option ok ? -i option to get mimetype
	gchar *s, *command = g_strdup_printf ("file -pr \"%s\"", filepath);
	E2_FILE *pipe = e2_fs_open_pipe (command);
	g_free (command);
	if (pipe == NULL)
	{
		*readable_type = NULL;
		*mime_type = NULL;
		return FALSE;
	}

#ifdef __USE_GNU
	ssize_t bytes_read = getdelim (&buf, &bsize, '\n', pipe);
	if (bytes_read >= 2)
	{
		*(buf+bytes_read-1) = '\0';
#else //getdelim() not available
	gchar *strtmp = fgets (buf, bsize, pipe);
	if (strtmp != NULL && *strtmp != '\n')
	{
		*(buf+strlen(buf)-1) = '\0';
#endif
		if ((s = strchr (buf, ':')) != NULL)	//if always ascii ':', don't need g_utf8_strchr()
			*readable_type = g_strdup (s+2);
		else
			*readable_type = g_strdup (buf);
	}
	else
	{
		*readable_type = NULL;
	}
	e2_fs_pipe_close (pipe);
	//get mimetype string
	command = g_strdup_printf ("file -bpri \"%s\"", filepath);
	pipe = e2_fs_open_pipe (command);
	g_free (command);
	if (pipe == NULL)
	{
		if (*readable_type != NULL)
		{
			g_free (*readable_type);
			*readable_type = NULL;
		}
		*mime_type = NULL;
		return FALSE;
	}

#ifdef __USE_GNU
	bytes_read = getdelim (&buf, &bsize, '\n', pipe);
	if (bytes_read >= 2)
	{
		*(buf+bytes_read-1) = '\0';
#else //getdelim() not available
	strtmp = fgets (buf, bsize, pipe);
	if (strtmp != NULL && *strtmp != '\n')
	{
		*(buf+strlen(buf)-1) = '\0';
#endif
		*mime_type = g_strdup (buf);
	}
	else
	{
		*mime_type = NULL;
	}
	e2_fs_pipe_close (pipe);
	g_free (buf);
	return TRUE;
}

/**
 * @brief dialog response callback
 *
 * @param dialog the permissions-dialog where the response was triggered
 * @param response the response for the clicked button
 * @param result pointer to store for the dialog return value
 *
 * @return
 **/
static void _e2_file_info_dialog_response_cb (GtkDialog *dialog, gint response,
	DialogButtons *result)
{
	gtk_widget_destroy (GTK_WIDGET (dialog));
	if (response == GTK_RESPONSE_CLOSE)	//not covered in the general decoder
	{
		*result = CANCEL;
		gtk_main_quit ();
	}
	else
		e2_dialog_response_decode_cb (dialog, response, result);
}

DialogButtons e2_file_info_dialog_run (gchar *localpath, gboolean multi)
{
	GtkWidget *dialog;
	GtkWidget *dialog_vbox;
	GtkWidget *table;
	GString *label_text = g_string_sized_new (NAME_MAX + 32);  //this is plenty big for all uses here
	gchar *s1, *s2;
	gchar perm_string[32];
	gchar date_string[32];
    struct stat statbuf;
	struct tm *tm_ptr;
	struct passwd *pw_buf;
	struct group *grp_buf;

	if (e2_fs_lstat (localpath, &statbuf))
		return CANCEL;

	//must prevent the default dialog response, it stuffs up Q cleanups
	DialogButtons choice;
	dialog = e2_dialog_create (NULL, NULL, _("file info"),
		_e2_file_info_dialog_response_cb, &choice);
	dialog_vbox = GTK_DIALOG (dialog)->vbox;
	gtk_container_set_border_width (GTK_CONTAINER (dialog_vbox), E2_PADDING);

	s1 = g_filename_display_basename (localpath);
	s2 = g_markup_escape_text (s1, -1);
	g_string_printf (label_text, "<b>%s</b>", s2);
	e2_widget_add_label (dialog_vbox, label_text->str, 0.5, TRUE, 0);
	g_free (s1);
	g_free (s2);
	e2_widget_add_separator (dialog_vbox, TRUE, E2_PADDING);

	//for regular files, get type-data from 'file' command
	if (S_ISREG (statbuf.st_mode))
	{
		gchar *mime = NULL, *readable = NULL;
		if (!_e2_file_info_dialog_get_file_type (localpath, &readable, &mime))
		{
			readable = NULL;	//ensure s2 etc not freed later
			s2 = _("Unknown");
		}
		else
			s2 = e2_utf8_from_locale (readable);	//freeme later

		GdkPixbuf *pixbuf = NULL;
		if (mime != NULL)
		{
			GtkIconTheme *icon_theme = gtk_icon_theme_get_for_screen
				(gtk_widget_get_screen (app.main_window));	//app.main_window->screen;
			//mime string returned by 'find' is of the form "text/plain; <stuff>"
			//from freedesktop.org, icon name is of the form media-subtype (e.g. text-plain).
			//If no specific icon is found, fallback to media-x-generic (e.g. text-x-generic).
			s1 = strchr (mime, ';');
			if (s1 != NULL)
				*s1 = '\0';
			s1 = strchr (mime, '/');
			if (s1 != NULL)
				*s1 = '-';
			pixbuf = gtk_icon_theme_load_icon (icon_theme, mime,
				GTK_ICON_SIZE_LARGE_TOOLBAR, 0, NULL);
			if (pixbuf == NULL)
			{
				if (s1 != NULL)
					*s1 = '\0';
				gchar *inm = g_strconcat (mime, "-x-generic", NULL);
				pixbuf = gtk_icon_theme_load_icon (icon_theme, inm,
					GTK_ICON_SIZE_LARGE_TOOLBAR, 0, NULL);
				g_free (inm);
				if (pixbuf == NULL)
				{	//fallback to anything that looks half-reasonable !
					s1++;	//pass the \0
					GList *inames = gtk_icon_theme_list_icons (icon_theme, NULL);
					GList *member;
					for (member=inames; member != NULL; member = member->next)
					{
						inm = (gchar *)member->data;
	//					printd (DEBUG, "candidate name = %s", inm);
						if (strstr (inm, s1) != NULL)
						{
							pixbuf = gtk_icon_theme_load_icon (icon_theme, inm,
								GTK_ICON_SIZE_LARGE_TOOLBAR, 0, NULL);
							break;
						}
					}
					e2_list_free_with_data (&inames);
				}
			}
		}
		if (pixbuf != NULL)
		{
			GtkWidget *image = gtk_image_new_from_pixbuf (pixbuf);
			GtkWidget *hbox = e2_widget_add_box (dialog_vbox, TRUE, 0, FALSE,
				FALSE, E2_PADDING_LARGE);
			gtk_box_pack_start (GTK_BOX (hbox), image, FALSE, FALSE, 0);
			//CHECKME escape label ?
			e2_widget_add_label (hbox, s2, 0, TRUE, 0);
			g_object_unref (G_OBJECT (pixbuf));
		}
		else
		{
			g_string_printf (label_text, "%s: %s", _("Type"), s2);
			//CHECKME escape label ?
			e2_widget_add_label (dialog_vbox, label_text->str, 0, TRUE, 0);
		}
		if (readable != NULL)
		{
			g_free (readable);
			g_free (mime);
			g_free (s2);
		}
	}
	//links to dirs reported as dirs by (e2_fs_is_dir3 (localpath))
	else if (S_ISDIR (statbuf.st_mode))
	{
		GtkIconTheme *icon_theme = gtk_icon_theme_get_for_screen
				(gtk_widget_get_screen (app.main_window));	//app.main_window->screen;
		GdkPixbuf *pixbuf = gtk_icon_theme_load_icon (icon_theme, "stock_open",
				GTK_ICON_SIZE_LARGE_TOOLBAR, 0, NULL);
		s2 =  _("Directory");
		if (pixbuf != NULL)
		{
			GtkWidget *image = gtk_image_new_from_pixbuf (pixbuf);
			GtkWidget *hbox = e2_widget_add_box (dialog_vbox, TRUE, 0, FALSE,
				FALSE, E2_PADDING_LARGE);
			gtk_box_pack_start (GTK_BOX (hbox), image, FALSE, FALSE, 0);
			//CHECKME escape label ?
			e2_widget_add_label (hbox, s2, 0, TRUE, 0);
			g_object_unref (G_OBJECT (pixbuf));
		}
		else
		{
			g_string_printf (label_text, "%s: %s", _("Type"), s2);
			e2_widget_add_label (dialog_vbox, label_text->str, 0, TRUE, 0);
		}
		//FIXME report the device holding the dir
	}
	else // not regular or directory
	{
		if (S_ISLNK (statbuf.st_mode))
			s2 = _("Symbolic Link");
		else if (S_ISCHR (statbuf.st_mode))
			s2 = _("Character Device");
		else if (S_ISBLK (statbuf.st_mode))
			s2 = _("Block Device");
		else if (S_ISFIFO (statbuf.st_mode))
			s2 = _("FIFO Pipe");
		else if (S_ISSOCK (statbuf.st_mode))
			s2 = _("Socket");
		else
			s2 = _("Unknown");

		g_string_printf (label_text, "%s: %s", _("Type"), s2);
		e2_widget_add_label (dialog_vbox, label_text->str, 0, TRUE, 0);
	}

	if (S_ISLNK (statbuf.st_mode))
	{
		gchar linkto[PATH_MAX+NAME_MAX];
		gchar *filename = g_strdup (localpath);
		gint len = strlen (filename);
		if (filename[len-1] == G_DIR_SEPARATOR)
			filename[len-1] = '\0';

		len = readlink (filename, linkto, sizeof (linkto));	//FIXME vfs
		if (len > 0)
		{
			linkto[len] = '\0';
			s1 = F_DISPLAYNAME_FROM_LOCALE (linkto);
			s2 = g_markup_escape_text (s1, -1);
			g_string_printf (label_text, " %s %s", _("to"), s2);
			F_FREE (s1);
			g_free (s2);
			//check the real target, without looking thru chained links
			struct stat statbuf2;	//don't disturb the size etc data for the link itelf
			if (e2_fs_lstat (linkto, &statbuf2) && errno == ENOENT)
				g_string_append_printf (label_text, "\n %s",_("(which is <b>missing</b>)"));

			e2_widget_add_label (dialog_vbox, label_text->str, 0, TRUE, 0);
		}
		else
			e2_widget_add_label (dialog_vbox, _("(Cannot resolve the link)"), 0, TRUE, 0);

		g_free (filename);
	}

	table = e2_widget_add_table (dialog_vbox, 9, 2, FALSE, TRUE, E2_PADDING);
	gtk_table_set_col_spacings (GTK_TABLE (table), E2_PADDING);

	e2_widget_add_label_to_table (table, _("Size:"), 0, 0,1,0,1);

	if (statbuf.st_size < 10240) // less than 10K
		g_string_printf (label_text, "%d %s", (gint) statbuf.st_size, _("bytes"));
	else if (statbuf.st_size < 1048576) // less than a meg
		g_string_printf (label_text, "%.2f %s", (gfloat) statbuf.st_size / 1024.0, _("kbytes"));
	else // more than a meg
		g_string_printf (label_text, "%.2f %s", (gfloat) statbuf.st_size / 1048576.0, _("Mbytes"));

	e2_widget_add_label_to_table (table, label_text->str, 0, 1,2,0,1);

	GtkWidget *sep = gtk_hseparator_new ();
	gtk_table_attach_defaults (GTK_TABLE (table), sep, 0,2,1,2);

	e2_widget_add_label_to_table (table, _("User:"), 0, 0,1,2,3);
	if ((pw_buf = getpwuid (statbuf.st_uid)) != NULL)
		g_string_printf (label_text, "%s", pw_buf->pw_name);
	else
		g_string_printf (label_text, "%d", (guint) statbuf.st_uid);
	e2_widget_add_label_to_table (table, label_text->str, 0, 1,2,2,3);

	e2_widget_add_label_to_table (table, _("Group:"), 0, 0,1,3,4);
	if ((grp_buf = getgrgid (statbuf.st_gid)) != NULL)
		g_string_printf (label_text, "%s", grp_buf->gr_name);
	else
		g_string_printf (label_text, "%d", (guint) statbuf.st_gid);
	e2_widget_add_label_to_table (table, label_text->str, 0, 1,2,3,4);
	e2_fs_get_perm_string (perm_string, sizeof (perm_string), statbuf.st_mode);
	g_string_assign (label_text, perm_string + 1);	//skip the "type code" in 1st byte of string
	g_string_insert_c (label_text, 6, ' ');
	g_string_insert_c (label_text, 3, ' ');
	e2_widget_add_label_to_table (table, _("Permissions:"), 0, 0,1,4,5);
	e2_widget_add_label_to_table (table, label_text->str, 0, 1,2,4,5);

	sep = gtk_hseparator_new ();
	gtk_table_attach_defaults (GTK_TABLE (table), sep, 0,2,5,6);

	s1 = "%a %Od %b %EY %OI:%OM %p";
	tm_ptr = localtime (&statbuf.st_atime);
	strftime (date_string, sizeof (date_string), s1, tm_ptr);
	e2_widget_add_label_to_table (table, _("Accessed:"), 0, 0,1,6,7);
	e2_widget_add_label_to_table (table, date_string, 0, 1,2,6,7);

	tm_ptr = localtime (&statbuf.st_mtime);
	strftime (date_string, sizeof (date_string), s1, tm_ptr);
	e2_widget_add_label_to_table (table, _("Modified:"), 0, 0,1,7,8);
	e2_widget_add_label_to_table (table, date_string, 0, 1,2,7,8);

	tm_ptr = localtime (&statbuf.st_ctime);
	strftime (date_string, sizeof (date_string), s1, tm_ptr);
	e2_widget_add_label_to_table (table, _("Changed:"), 0, 0,1,8,9);
	e2_widget_add_label_to_table (table, date_string, 0, 1,2,8,9);

	g_string_free (label_text,TRUE);

	if (multi)
	{
		E2_BUTTON_NOTOALL.showflags &= ~!E2_BTN_DEFAULT;
		e2_dialog_add_defined_button (dialog, &E2_BUTTON_NOTOALL);
		e2_dialog_set_negative_response (dialog, E2_RESPONSE_NOTOALL);
	}
	E2_BUTTON_CLOSE.showflags |= E2_BTN_DEFAULT;
	e2_dialog_add_defined_button (dialog, &E2_BUTTON_CLOSE);

	e2_dialog_setup (dialog, app.main_window);

#ifdef E2_ASYNC
	gdk_threads_enter ();
#endif
	e2_dialog_run (dialog, app.main_window, 0);
	gtk_window_present (GTK_WINDOW (dialog));	//sometimes, dialog is not focused
	gtk_main ();
#ifdef E2_ASYNC
	gdk_threads_leave ();
#endif
	return choice;
}
